%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

%
% psDQ implements heat transfer functions:
%    1. Woschni, see Heywood, 1988
%    2. Hohenberg, see SAE 7900825
%
% 
% Parameters:
%   B       [m]     bore
%   S       [m]     stroke
%   Tw      [K]     wall temperature
%   C1, C2  [-]     weights for characteristic speed due to piston motion
%                   and combusition
%   --> Parameters are fields of global variables of type structure. 
%  
% Input:
%   pm      [Pa]    motored pressure
%   pf      [Pa]    fired pressure
%   T       [K]     Gas temperature
%   A       [m^2]   instantaneous contact area between gas and cylinder wall
%   V       [m^3]   instantaneous cylinder volume
%   pivc    [Pa]    pressure at intake valve close
%   Tivc    [k]     temperature at intake valve close
%   Vivc    [m^3]   volume at intake valve close
%   Ts      [K]     Temperature of cylinder wall surface
% 
% Output:
%   DQ      [J/rad] heat transfer to the wall
%   h       [W/(K*m^2)] heat transfer coefficient
% 
% Syntax:   [DQ, h] = psDQ(pm, pf, T, V, w, pivc, Tivc, Vivc,Ts)
function [DQ, h] = psDQ(pm, pf, T, V, A, mps, pivc, Tivc, Vivc, Ts)
  
global psSimStruct
  
B  = psSimStruct.psGeometry.B;
S  = psSimStruct.psGeometry.S;
Vd = pi*B^2/4*S;
Tw = Ts;
C1 = psSimStruct.psHeat.C1;
C2 = psSimStruct.psHeat.C2;
    
% convert from SI dimensions into Woschni's dimensions   
pfa	= pf/101325; % [Pa] --> [atm]    
% heat transfer coefficient:
charW = 2.28*((0.0034*(pf-pm)*Tivc*Vd*C2)/(pivc*Vivc)+mps);
h = 131*C1*B^-0.2*pfa^0.8*charW^0.8*T^-0.53;    
%DQ = zeros(size(A*h*(T-Tw)));   
DQ = A*h*(T-Tw);


%case 'Hohenberg'
%B  = psSimStruct.psGeometry.Bore;
%S  = psSimStruct.psGeometry.Stroke;
%Tw = Ts;
%C1 = psSimStruct.psHeat.C1;
%C2 = psSimStruct.psHeat.C2;   
%% convert from SI dimensions into Woschni's dimensions   
%pf		= pf*1e-5; % [Pa] --> [bar]
%% heat transfer coefficient:
%h	= C1   *   V^-0.06   *   pf^0.8   *   T^-0.4  * (mps + C2)^0.8;    
%DQ	= A*h*(T-Tw);
