%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function [dXpVT dXgc] = psDX(XpVT,Xgc,Xgc_to,dV,dm,dQ,T_ext,Xgc_ext,Xgc_ext_to,dm_ext)
% function dX = psDX(XpVT,Xgc,dV,dm,dQ,extFlow);
%    XpVT    = [p, V1, T1 , V2 , T2, ... ]'   (1+2*N)x1
%    Xgc     = [x_uf x_ua x_bf  x_ba]         NxK matrix
%    Xgc_to  = ... same but takes into account the destination Xgc-part
%    dV      = Total volume change            Scalar
%    dm      = Mass flow between zones        N x N matrix 
%    dQ      = Heat transfer from resp zone   Nx1
%  External mass flow: There may only be one flow upstreams and one flow
%  downstreams to each zone => Two flows for each zone
%    Text    = Temperture outside the control volume  Nx2
%    Xgc_ext = Gas composition upstreams              NxKx2
%    Xgc_ext_to = ... same but takes into account the destination Xgc-part
%    dm_ext  = Mass flow TO the control volume        Nx2 
%      * Note the sign/direction convention
% Output:
%    dXpVT = (1+2*N)x1
%    dXgc  = NxK
%
%
%
% OVERVIEW OF dm :
%   _________________
%   I dm11 ... dm1n I  To zone 1
%   I dm21 ... dm2n I          2
%   I   .  .    .   I          .
%   I   .   .   .   I          .
%   I   .     . .   I          .
%   I dmn1 ... dmnn I          n
%   -----------------
%      1   ...   n
%   From zone
%
% This means:
%  * dm(i,j) > 0 => gas is entering zone i from zone j
%  * dm(i,j) < 0 => gas is leaving zone i for zone j 
%
%
% Author: Per Öberg (renewed implementation of old psPack with multiple
%                    authors)
%
% Changelog: 
%  2007-12-17 - Created

zones = (length(XpVT)-1)/2;
global psSimStruct
K = psSimStruct.psThermProp.K;

% Define index vecetors for the different states
% and use only these to avoid indexation mistakes.
tIdx = (1:zones)*2+1; 
vIdx = (1:zones)*2;

% No data in X can be <0, physical limits for volume,
% pressure and temperature. This could generate a warning but since
% this will most probably only happen for calculations that will be
% discarded it does not seem necessary.

% Cylinder pressure is element 1 of input X
p = XpVT(1);  
T = XpVT(tIdx);
V = XpVT(vIdx);

V(find(V<0)) = 0;

if p <= 0 || min(T ) <= 0 || min(V) < 0
    disp('Warning p, V or T out of bounds')
    %dXpVT = NaN*ones(size(XpVT));
    %dXgc = NaN*ones(size(Xgc));
    %return
end

      

% This is where we get the thermodynamic properties
% We need:
% 1) For each zone everything is needed 
outputs = ones(1,9);
[h cp cv R dRdp dRdT dudp dudX dRdX]=psThermProp(p,T,Xgc,outputs);


% 2) for external zones only those who are from in-flows enthalpy is
% realy needed. For now, however, the implementation calculates 
% everything no matter if it's already been calulated before.
T_hat   = (dm_ext <= 0 ).*repmat(T,1,2)+(dm_ext > 0 ).*T_ext;
Xgc_hat(:,:,1) = repmat((dm_ext(:,1) > 0),1,K).*Xgc_ext(:,:,1)+...
                 repmat((dm_ext(:,1) <= 0),1,K).*Xgc;
Xgc_hat(:,:,2) = repmat((dm_ext(:,2) > 0),1,K).*Xgc_ext(:,:,2)+...
                 repmat((dm_ext(:,2) <= 0),1,K).*Xgc;

Xgc_hat_to(:,:,1) = repmat((dm_ext(:,1) > 0),1,K).*Xgc_ext_to(:,:,1)+...
                 repmat((dm_ext(:,1) <= 0),1,K).*Xgc;
Xgc_hat_to(:,:,2) = repmat((dm_ext(:,2) > 0),1,K).*Xgc_ext_to(:,:,2)+...
                 repmat((dm_ext(:,2) <= 0),1,K).*Xgc;


outputs = zeros(1,9);
outputs(1) = 1;
h_hat(:,1) = psThermProp(p,T_hat(:,1),Xgc_hat(:,:,1),outputs);
h_hat(:,2) = psThermProp(p,T_hat(:,2),Xgc_hat(:,:,2),outputs);

% Calculate dXgc as soon as possible
mdXgc = zeros(zones,K);
for ii = 1:zones
    % Calculate m*dXgc for zone ii
    mdXgc(ii,:) = (dm(ii,:).*(dm(ii,:)>0))*(Xgc_to-repmat(Xgc(ii,:),zones,1));
end
mdXgc=mdXgc+(Xgc_hat_to(:,:,1)-Xgc).*repmat((dm_ext(:,1).*(dm_ext(:,1)>0)),1,K);
mdXgc=mdXgc+(Xgc_hat_to(:,:,2)-Xgc).*repmat((dm_ext(:,2).*(dm_ext(:,2)>0)),1,K);
%dXgc=mdXgc/m;
Veps = V; 
Veps(find(V./sum(V) < 10*eps)) = eps;
dXgc=repmat(R.*T./(p*Veps),1,K).*mdXgc;
dXgc(find( sum(isnan(dXgc),2)),:) = 0;

%disp('Warning: There has to be a more efficient way!')
%if any(dXgc(:))
%    disp('Debug: in psDX.m We have flow')
%end

%dXgc = dXgc';
if ~isreal(sum(dXgc))
    disp('Error in psDX.m: Complex dXgc')
end


% For internal flows we have everything we need by now. We only need to
% structure it a bit.
%  h_hat_int (internal) is the enthalpy of the gas that enters or leaves
%  the zones internally
%  * dm(i,j) > 0 => gas is entering zone i from zone j => h_hat_int(i,j) = h(j)
%  * dm(i,j) < 0 => gas is leaving zone i for zone j => h_hat_int(i,j) = h(i)
%  * dm(i,i) = 0 sinze we have no mass tranfer to ourselfes.
%  * h(i,i) = hi(i) (But could be set to zero).
%  * hij is thus symmetric, but dm is antisymmetric.
%
% Remember that h is a vector with entalpies.
% The right elements of h shall now be placed at the rigt positions in 
% h_hat_int. We can use the fact that depending on the sign of dm(i,j) we shold
% pick either [h ... h](i,j) or [h' ; ... ; h'](i,j). Since h_hat_int is
% symmetric we can take those elements of [h ... h] where dm is
% positive and put them in h_hat_int. We can then transpose h_hat_int and
% add it to itself. After that only those elements in h_hat_int that
% corresponds to  dm(i,j)=0 is missing (e.g. the diagonal of h_hat_int.)
h_hat_int = (sign(dm) == 1 ).*(repmat(h',zones,1));
h_hat_int = h_hat_int + h_hat_int';%+diag(hi) (Removed to save comuptation time since it is
                 %unused)
% Note: h_hat_int(i,j)=0 for all elements that correpsonds to dm(i,j)=0
% after this operation. However, as will be seen later 
% h_hat_int for elements with dm(i,j)=0 does not enter the equation at all.



% Build the matrix components and then finally the A-matrix. Note that it's
% the same for both "well stirred mixer" and "Well stirred reactor" since
% dRdp=dRdT=dudp=0 in the well stirred mixer case. The matrix is shifted
% for easier implementation

m = p.*V./R./T;
a_m = (R.*T./p-T.*dRdp);
a = a_m.*m;
b_m =-(R+T.*dRdT);
c_m =dudp;
d_m = cv;

Bshift = [dV ;...
          R.*T.*(sum(dm_ext,2)+sum(dm,2))+...
          dot(dRdX,mdXgc,2).*T;...
          -dQ-dot(dudX,mdXgc,2)+...
          (h_hat(:,1)-h+R.*T).*dm_ext(:,1)+...
          (h_hat(:,2)-h+R.*T).*dm_ext(:,2)+...
          sum((h_hat_int-repmat(h-R.*T,1,zones)).*dm,2)];

% 1) Shifting the rows in B is equal to shifting the rows of A
% 2) Shifting the columns of A is equal to shifting the rows in the dx
% vector. Thus for easy implementation we use the notation
% Ashift = 
% 0  1  ... 1  0  ... 0
% a1 p         b1	 
% .    .          .	 
% .      .          .	 
% .        .          .	 
% an        p          bn
% c1  p         d1	 
%  .    .          .	 
%  .      .          .	 
%  .        .          .	 
% cn         p          dn
% 
% etc...

%if V(1) > sum(V)*1e-10
%    disp('Whoho')
%end

zoneIndx = find(m./sum(m) > psSimStruct.psSimPar.minVFrac);
badzoneIndx = find(m./sum(m) <= psSimStruct.psSimPar.minVFrac);
%zoneIndx = 2
%badzoneIndx = 1
goodzones = length(zoneIndx);
badzones = length(badzoneIndx);


ADshift = [0 a(zoneIndx)'/sum(V) zeros(1,goodzones);...
           ones(goodzones,1)  eye(goodzones) diag(b_m(zoneIndx)./d_m(zoneIndx));...
           c_m(zoneIndx)./a_m(zoneIndx)       eye(goodzones) eye(goodzones)];

%shiftIdx = [1 2:2:goodzones*2 3:2:(goodzones*2+1)]';

D2shiftinv = diag([1; a(zoneIndx)/p; a_m(zoneIndx)./d_m(zoneIndx)]);
D1shiftinv = diag([p/sum(V); 1./a(zoneIndx); 1./a(zoneIndx)]);


%dXpVT=D2shiftinv(shiftIdx,shiftIdx)*(ADshift(shiftIdx,shiftIdx)\(D1shiftinv(shiftIdx,shiftIdx)*Bshift(shiftIdx)));


dXpVTgoodshift = D2shiftinv*(ADshift\(D1shiftinv*Bshift([1; zoneIndx+1; zones+zoneIndx+1])));
dXpVTbad(1:badzones,1)=R(badzoneIndx,:).*T(badzoneIndx,:).*(sum(dm_ext(badzoneIndx,:),2)+sum(dm(badzoneIndx,:),2))./p;
dXpVTbad(badzones+(1:badzones),1)=0;


if 0 % Partly working for calculating new zone temperature. Do not truse
     % as is.
    global zoneHasBeenBad
    if isempty( zoneHasBeenBad)
         zoneHasBeenBad= 0;
    end
    if ~isempty(badzoneIndx) && ...
            badzoneIndx == 1 && ...
            (sum(dm_ext(badzoneIndx,:),2)+sum(dm(badzoneIndx,:),2)) > 0
        zoneHasBeenBad = 1;
    end
    if zoneHasBeenBad == 1 && ...
            isempty(badzoneIndx)
        h_u = h(2);
        Tb = T(1);
        deltaT = Tb;
        lpCount = 0;
        while abs(deltaT) > 1e-10*T(1) && lpCount < 100
            [h_b cp_b] = psThermProp(p,Tb,Xgc(1,:),[1 1 0 0 0 0 0 0 0]);
            deltaT  = -(h_b-h_u)/cp_b;
            Tb = deltaT+Tb;
            lpCount = lpCount+1;
        end   
        zoneHasBeenBad = 0;
        disp(sprintf('Init Temperature should be %d',Tb))
    end
end

%if ~isempty(badzoneIndx)
  dXpVT([1; 2*zoneIndx; 1+2*zoneIndx; 2*badzoneIndx; 2*badzoneIndx+1],1) = [dXpVTgoodshift; dXpVTbad];
  % else
  %  dXpVT([1 1+zoneIndx ; zones+1+zoneIndx],1) = [dXpVTgood; dXpVTbad];
  % end

if ~isreal(sum(dXpVT))
    disp('Error in psDX.m: Complex dXpVT')
end

if any(isnan(dXgc(:))) || any(isinf(dXgc(:))) 
    disp('dXgc inf or nan')
end

if any(isnan(dXpVT)) || any(isinf(dXpVT)) 
    disp('dXpVT inf or nan')
end

