%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function ps1ZoneOpenCylinder(block)
setup(block);
%endfunction

%% Function: setup ===================================================
%% Abstract:
%%   Set up the S-function block's basic characteristics such as:
%%   - Input ports
%%   - Output ports
%%   - Dialog parameters
%%   - Options
%% 
%%   Required         : Yes
%%   C-Mex counterpart: mdlInitializeSizes
%%
function setup(block)
  %disp(sprintf('Running ps2ZoneOpenCylinder:Setup at time: %d',NaN))

  global psSimStruct
  K = psSimStruct.psThermProp.K;
  
  % Register number of ports
  block.NumInputPorts  = 4;
  block.NumOutputPorts = 3;

  
  % In signals: 
  % 1) theta
  % 2) inUp:
  %   a) dmUp
  %   b) pUp,TUp
  %   c) Xgc_Up
  % 3) inDown
  %   a) dmDown
  %   b) pDown,TDown
  %   c) Xgc_Down
  % 4) Ts (T_Wall)
  % Out signals:
  % 1) outUp (2+K)
  %   a) pUp,TUp
  %   b) Xgc_Up
  % 2) DiagOut (Cont states and dQ) (3+K+1)
  % 3) outDown (2+K)
  %   a) pDown,TDown
  %   b) Xgc_Down

  % Cont states
  %  X = [ p,V,T,-Xgc-]
  % Disc States:
  %   1) mdlStartTime
  
  % Update sker efter output i varje steg. Dvs saker händer i följande ordning:
  % 1) Setup
  % 2) SetInPortFrameData
  % 3) DoPostPropSetup
  % 4) Start (time exists from hereon)
  % 5) Init 
  % 6) Output
  % 7) Update
  % 8) Repeat 6-7 and possibly 5-7
  % Start kan alltså anvädas första ggn modellen startas för att
  % avgöra vars vi är  medans initialize kan köras varje gång vi kommer
  % tillbaks från ett disablat tillstånd.

  % None of this dynamic inport stuff
  block.SetPreCompPortInfoToDefaults;
     
  %%%%%%%%%%%%%%%%
  % Input ports
  %%%%%%%%%%%%%%%%
  % Data type
  block.InputPort(1).DatatypeID  = 0;  % double
  block.InputPort(2).DatatypeID  = 0;  % double
  block.InputPort(3).DatatypeID  = 0;  % double
  block.InputPort(4).DatatypeID  = 0;  % double

  % No complex signals
  block.InputPort(1).Complexity  = 'Real';
  block.InputPort(2).Complexity  = 'Real';
  block.InputPort(3).Complexity  = 'Real';
  block.InputPort(4).Complexity  = 'Real';

  % Dimensions are fixed
  block.InputPort(1).Dimensions  = 1;
  block.InputPort(2).Dimensions  = 3+K;
  block.InputPort(3).Dimensions  = 3+K;
  block.InputPort(4).Dimensions  = 1;

  % No direct feedthrough
  block.InputPort(1).DirectFeedthrough = false;
  block.InputPort(2).DirectFeedthrough = false;
  block.InputPort(3).DirectFeedthrough = false;
  block.InputPort(4).DirectFeedthrough = false;
  
  % Never accept frame data... 
  block.InputPort(1).SamplingMode  = 0;
  block.InputPort(2).SamplingMode  = 0;
  block.InputPort(3).SamplingMode  = 0;
  block.InputPort(4).SamplingMode  = 0;

  %%%%%%%%%%%%%%%%
  % Output ports
  %%%%%%%%%%%%%%%%  
  % Data type
  block.OutputPort(1).DatatypeID  = 0; % double
  block.OutputPort(2).DatatypeID  = 0; % double
  block.OutputPort(3).DatatypeID  = 0; % double

  % No complex signals
  block.OutputPort(1).Complexity  = 'Real';
  block.OutputPort(2).Complexity  = 'Real';
  block.OutputPort(3).Complexity  = 'Real';
  
  % Dimensions are fixed
  block.OutputPort(1).Dimensions  = 2+K;
  block.OutputPort(2).Dimensions  = 3+K+1;
  block.OutputPort(3).Dimensions  = 2+K;

  % Never accept frame data... 
  block.OutputPort(1).SamplingMode  = 0;
  block.OutputPort(2).SamplingMode  = 0;
  block.OutputPort(3).SamplingMode  = 0;
  
  % Register parameters
  block.NumDialogPrms     = 2;
  %  block.DialogPrmsTunable = {'Tunable','Nontunable','SimOnlyTunable'};

  % Register sample times
  %  [0 offset]            : Continuous sample time
  %  [positive_num offset] : Discrete sample time
  %
  %  [-1, 0]               : Port-based sample time
  %  [-2, 0]               : Variable sample time
  block.SampleTimes = [0 0];
  
  % Cont States are:
  %  X = [ p,V,T,-Xgc-]
  block.NumContStates = 3+K;

    
  %% -----------------------------------------------------------------
  %% Options
  %% -----------------------------------------------------------------
  % Specify if Accelerator should use TLC or call back into 
  % M-file
  block.SetAccelRunOnTLC(false);
  
  %% -----------------------------------------------------------------
  %% Register methods called during update diagram/compilation
  %% -----------------------------------------------------------------
  
  %% 
  %% CheckParameters:
  %%   Functionality    : Called in order to allow validation of
  %%                      block's dialog parameters. User is 
  %%                      responsible for calling this method
  %%                      explicitly at the start of the setup method
  %%   C-Mex counterpart: mdlCheckParameters
  %%
  block.RegBlockMethod('CheckParameters', @CheckPrms);

  %%
  %% PostPropagationSetup:
  %%   Functionality    : Setup work areas and state variables. Can
  %%                      also register run-time methods here
  %%   C-Mex counterpart: mdlSetWorkWidths
  %%
  block.RegBlockMethod('PostPropagationSetup', @DoPostPropSetup);

  %% -----------------------------------------------------------------
  %% Register methods called at run-time
  %% -----------------------------------------------------------------
  
  %% 
  %% ProcessParameters:
  %%   Functionality    : Called in order to allow update of run-time
  %%                      parameters
  %%   C-Mex counterpart: mdlProcessParameters
  %%  
  block.RegBlockMethod('ProcessParameters', @ProcessPrms);

  %% 
  %% InitializeConditions:
  %%   Functionality    : Called in order to initalize state and work
  %%                      area values
  %%   C-Mex counterpart: mdlInitializeConditions
  %% 
  block.RegBlockMethod('InitializeConditions', @InitializeConditions);
  
  %% 
  %% Start:
  %%   Functionality    : Called in order to initalize state and work
  %%                      area values
  %%   C-Mex counterpart: mdlStart
  %%
  block.RegBlockMethod('Start', @Start);

  %% 
  %% Outputs:
  %%   Functionality    : Called to generate block outputs in
  %%                      simulation step
  %%   C-Mex counterpart: mdlOutputs
  %%
  block.RegBlockMethod('Outputs', @Outputs);

  %% 
  %% Update:
  %%   Functionality    : Called to update discrete states
  %%                      during simulation step
  %%   C-Mex counterpart: mdlUpdate
  %%
  block.RegBlockMethod('Update', @Update);

  %% 
  %% Derivatives:
  %%   Functionality    : Called to update derivatives of
  %%                      continuous states during simulation step
  %%   C-Mex counterpart: mdlDerivatives
  %%
  block.RegBlockMethod('Derivatives', @Derivatives);
  

  %% 
  %% ZeroCrossings:
  %%   Functionality    : For S-functions with continuous sample
  %%                      time, if input signals have discontinuities,
  %%                      this method is called to detect 
  %%                      discontinuities
  %%   C-Mex counterpart: mdlZeroCrossings
  %%
  block.RegBlockMethod('ZeroCrossings', @ZeroCrosssings);
  
  %% 
  %% SimStatusChange:
  %%   Functionality    : Called when simulation goes to pause mode
  %%                      or continnues from pause mode
  %%   C-Mex counterpart: mdlSimStatusChange
  %%
  block.RegBlockMethod('SimStatusChange', @SimStatusChange);
  
  %% 
  %% Terminate:
  %%   Functionality    : Called at the end of simulation for cleanup
  %%   C-Mex counterpart: mdlTerminate
  %%
  block.RegBlockMethod('Terminate', @Terminate);

  %% -----------------------------------------------------------------
  %% Register methods called during code generation
  %% -----------------------------------------------------------------
  
  %%
  %% WriteRTW:
  %%   Functionality    : Write specific information to RTW file
  %%   C-Mex counterpart: mdlRTW
  %%
  block.RegBlockMethod('WriteRTW', @WriteRTW);

%endfunction

%% -------------------------------------------------------------------
%% The local functions below are provided for illustrative purposes
%% to show how you may implement the various block methods listed
%% above.
%% -------------------------------------------------------------------

function CheckPrms(block)
    %disp(sprintf('Running ps2ZoneOpenCylinder:CheckPrms at time: %d',NaN))
%  a = block.DialogPrm(1).Data;
%  if ~strcmp(class(a), 'double')
%    error('Invalid parameter');
%  end
  
%endfunction

function ProcessPrms(block)
  %disp(sprintf('Running ps2ZoneOpenCylinder:ProcessPrms at time: %d',block.currentTime))
  block.AutoUpdateRuntimePrms;
 
%endfunction
  
function DoPostPropSetup(block)
  %disp(sprintf('Running ps2ZoneOpenCylinder:DoPostPropSetup at time: %d',NaN))

  % Disc States:
  %   1) mdlStartTime
 
  block.NumDworks = 2;
  block.Dwork(1).Name            = 'mdlStartTime';
  block.Dwork(1).Dimensions      = 1;
  block.Dwork(1).DatatypeID      = 0;      % double
  block.Dwork(1).Complexity      = 'Real'; % real
  block.Dwork(1).UsedAsDiscState = true;

  block.Dwork(2).Name            = 'debugFlag'; % To help debug
  block.Dwork(2).Dimensions      = 1;
  block.Dwork(2).DatatypeID      = 0;      % double
  block.Dwork(2).Complexity      = 'Real'; % real
  block.Dwork(2).UsedAsDiscState = true;

  %% Register all tunable parameters as runtime parameters.
  block.AutoRegRuntimePrms;

%endfunction

function InitializeConditions(block)
  global psSharedSimData
  global psSimStruct
  CylNbr = block.DialogPrm(2).Data;
  %disp(sprintf('Running ps2ZoneOpenCylinder:InitConditions at time: %d',...
  %             block.currentTime));
  caOffset = block.DialogPrm(1).Data;
  currentCa = block.currentTime*psSimStruct.psSimPar.w+caOffset;
  if block.currentTime == block.Dwork(1).Data
      if 0 
          disp('Initconditions in ps2ZoneOpenCylinder')
      end
      initStates = psInitGasExchange(currentCa);
      block.ContStates.Data = initStates;
  else        
      %warning('ps2ZoneOpenCylinder:InitConditions',...
      %        'Check ordering of states in ps2ZoneCylinder');
      % Xend = [ p,Vb,Tb,Vu,Tu,-Xgc_b-,-Xgc_u-]
      Xend = [psSharedSimData.CombBlock(CylNbr).ContStates.Data;];        
      initStates = psInitGasExchange(currentCa,Xend);
      block.ContStates.Data = initStates;
  end
%endfunction

function Start(block)
  global psSharedSimData
  CylNbr = block.DialogPrm(2).Data;
  %disp(sprintf('Running ps2ZoneOpenCylinder:Start at time: %d',block.currentTime))
  % Save start time in Dwork(3)  to be able to use it for initializing later
  block.Dwork(1).Data = block.currentTime;
  % Debug flag  is zero
  block.Dwork(2).Data = 0;
  % Make sure that the block is saved in the shared simdata
  psSharedSimData.GasExchBlock(CylNbr) = block;
%endfunction

function WriteRTW(block)
   block.WriteRTWParam('matrix', 'M',    [1 2; 3 4]);
   block.WriteRTWParam('string', 'Mode', 'Auto');
   
%endfunction

function Outputs(block)
%  disp(sprintf('Running ps2ZoneOpenCylinder:Outputs at time: %d',...
%                block.currentTime))
  
  global psSimStruct
  w = psSimStruct.psSimPar.w;
  K = psSimStruct.psThermProp.K;
  S = psSimStruct.psGeometry.S;

  % Get the in signals in the old form
  % 1) theta
  theta = block.InputPort(1).Data(1);
  % 4) Ts (T_Wall)
  Ts =block.InputPort(4).Data(1);

  % Cont. States:
  %   p,V,T,-Xgc-
  X  = block.ContStates.Data;
  p  = X(1);
  T  = X(3);
%   if block.IsMajorTimeStep
%       if p < 1e5
%           disp('ps1ZoneCylinder: Setting Debug flag since p<1e5')
%           disp(sprintf('Time: %f is major',block.CurrentTime));
%           block.Dwork(2).Data = 1;
%       end
%   end

  
  % Define heat transfer out of the zones
  %  V = psVolume(theta);
  A = psArea(theta);
  mps = S*w/pi;
  dQ = psDQlp(p,T,A,mps,Ts);

  if ~isreal(sum(dQ))
      % disp('dQ un-Real')

      % This means that the there is something wrong. Most
      % likely it's the tempearature that is negative. This occurs sometimes
      % when simulink tries to take a very large step. Therefore , if xB is
      % small dQ could be set to zeero to avoid this problem.
      dQ = 0;
  end
  
  % In signals: 
  % 1) theta
  % 2) inUp:
  %   a) dmUp
  %   b) pUp,TUp
  %   c) Xgc_Up
  % 3) inDown
  %   a) dmDown
  %   b) pDown,TDown
  %   c) Xgc_Down
  % 4) Ts (T_Wall)
  % Out signals:
  % 1) outUp (2+K)
  %   a) pUp,TUp
  %   b) Xgc_Up
  % 2) DiagOut (Cont states and dQ) (3+K+1)
  % 3) outDown (2+K)
  %   a) pDown,TDown
  %   b) Xgc_Down

  % Cont states
  %  X = [ p,V,T,-Xgc-]
  % Disc States:
  %   1) mdlStartTime

  %  if block.IsMajorTimeStep
  %      disp(sprintf('Time: %f is major',block.CurrentTime));
  %  else
  %      disp(sprintf('Time: %f is minor',block.CurrentTime));
  %  end
  
  % Upstreams
  block.OutputPort(1).Data = block.ContStates.Data([1 3 3+(1:K)]);
  % Diag port
  block.OutputPort(2).Data = [block.ContStates.Data ; sum(dQ)]; 
  % Downstreams
  block.OutputPort(3).Data = block.ContStates.Data([1 3 3+(1:K)]);
%endfunction

function Update(block)
% disp(sprintf('Running ps2ZoneOpenCylinder:Update at time: %d',block.currentTime))
%endfunction

function Derivatives(block)
%  disp(sprintf('Running ps2ZoneOpenCylinder:Derivatives at time: %d',block.currentTime))
  % Get/Set parameters:

  global psSimStruct
  K = psSimStruct.psThermProp.K;
  global psSimData

  w = psSimStruct.psSimPar.w;
  S = psSimStruct.psGeometry.S;

  % Get the in signals in the old form
  % 1) theta
  theta = block.InputPort(1).Data(1);
  % 2) inUp:
  %   a) dmUp
  dmUp       = block.InputPort(2).Data(1);
  %   b) pUp,TUp
  TUp        = block.InputPort(2).Data(3);
  %   c) Xgc_FlUp
  Xgc_FlUp    = block.InputPort(2).Data(3+(1:K))';
  % 3) inDown
  %   a) dmDown
  dmDown       = block.InputPort(3).Data(1);
  %   b) pDown,TDown
  TDown        = block.InputPort(3).Data(3);
  %   c) phiFlDown,phiResFlDown,xResFlDown
  Xgc_FlDown    = block.InputPort(3).Data(3+(1:K))';

  T_ext      = [TUp TDown];
  Xgc_ext(1,:,:)  = [Xgc_FlUp'  Xgc_FlDown'];
  Xgc_ext_to = Xgc_ext;
  dm_ext     = [dmUp -dmDown];
  
  
  % 4) Ts (T_Wall)
  Ts =block.InputPort(4).Data(1);

  % Cont. States:
  %   p,V,T,-Xgc-
  %    -  Select the p,V,T states
  X  = block.ContStates.Data;
  XpVT = X(1:3)';
  Xgc = X(4:end)';
  p  = XpVT(1);
  T  = XpVT(3);
  
%   if block.Dwork(2).Data
%       disp('ps1ZoneCylinder: Debug flag set!')
%   end
  
  
  % Define massflow to and from the zones   
  % dm(i,j) is the massflow from zone j to zone i.
  % Massflows entering zone i have a positive sign.
  % dm(i,i) are massflows entering the cylinder from
  % outside.
  dm = 0; 

  % Define total derivative of total cylinder Volume
  dV = w*psDVolume(theta);

  % Define heat transfer out of the zones
  %  V = psVolume(theta);
  A = psArea(theta);
  mps = S*w/pi;
  dQ = 0*psDQlp(p,T,A,mps,Ts);

  
  [dXpVT dXgc]=  psDX(XpVT,Xgc,zeros(1,K),dV,dm,dQ,T_ext,Xgc_ext,Xgc_ext_to,dm_ext);  

  % Cont states
  %  X = [ p,V,T,-Xgc-]
  if ~isreal(sum(dXpVT))|| ~isreal(sum(dXgc))
      % This means that the there is something wrong. Most likely it's
      % the tempearature that is negative, or the motored pressure that
      % is larger than the real pressure. This occurs sometimes for very
      % small zones when simulink tries to take a very large
      % step. Therefore  The values could be set to NaN so that Simulink
      % detects the error and takes a shorter step
      dXpVT = NaN(size(dXpVT));
      dXgc  = NaN(size(dXgc));
  end
  block.Derivatives.Data = [dXpVT(:) ;dXgc(:)];

%endfunction

function ZeroCrosssings(block)
  % Eventuellt lägga till lite kollar för zero-cross hjälp.
  % Tillstånden är:
  % - p,Vbg,Tbg,Vfm,Tfm,Tav,phi_bg,phiRes_bg,xRes_bg,phi_fm,phiRes_fm,xRes_fm
  % och det som kan hända med dem är att Volymerna kan närma sig gränsen
  % för vad som är acceptabel storlek. De viktigaste eventsen är redan
  % omhändertagna men även events som ändring av riktning på flöden mm
  % är av intresse. Ett bra sätt är att kolla efter if-satser i
  % derivata-uträkningen och sätta zero-cross-detection på signalerna i
  % omgivningen som har med signalen att göra, om det är viktigt vill säga.
%endfunction

function SimStatusChange(block, s)
  
  if s == 0
    disp('Pause has been called');
  elseif s == 1
    disp('Continue has been called');
  end
  
%endfunction
    
function Terminate(block)
%endfunction
 
