%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function psMultiGasPipe(block)
 
setup(block);
  
function setup(block)
  global psSimStruct
  K = psSimStruct.psThermProp.K;
  %% Register number of dialog parameters   
  block.NumDialogPrms = 10;
  % 01) N             Number of parts/zones
  % 02) NumInputs     Number of inputs
  % 03) x             Pipe Length
  % 04) A             Pipe Area
  % 05) Pinit         Initial pressure
  % 06) Tinit         Initial temperature
  % 07) Xgc_ini       Initial Xc
  % 08) kFric         Friction coefficient 
  % 09) connectorUp   Upstreams connector is Restr (= 1) or CV (= 2)
  % 10) connectorDown Downstreams connector is Restr (= 1) or CV (= 2)
  

  %% Setup States
  % N parts gives N pressures, N temperatures, N*K Xgc and
  % N+1 possible speeds (depending on upstreams connectors).
  %
  % The states are: 
  %  - p=(1:N), T=(1:N)+N, Xc=(1:(N*K))+2*N, U =(1:(N+1))+2*N+N*K; 
  block.NumContStates = block.DialogPrm(1).Data*(3+K)+1; 

  
  %% Register number of input and output ports
  block.NumInputPorts  = 2;
  block.NumOutputPorts = 3;

  %% Setup functional port properties to dynamically
  %% inherited.
  %block.SetPreCompInpPortInfoToDynamic;
  %block.SetPreCompOutPortInfoToDynamic;
  block.SetPreCompPortInfoToDefaults;
 
  
  % Input dimenstions: NumInputs*(mDot, state (p,T),Xc)
  block.InputPort(1).Dimensions = (3+K)*block.DialogPrm(2).Data(1);
  block.InputPort(2).Dimensions = (3+K)*block.DialogPrm(2).Data(2);

  for ii = 1:2
      block.InputPort(ii).DirectFeedthrough = false;
  end
 
  % Output dimensions: 
  %   1 ) mDot, p, T, Xc
  block.OutputPort(01).Dimensions = 3+K; 
  %   2)  p=(1:N), T=(1:N)+N, Xc=(1:(N*K))+2*N, U =(1:(N+1))+2*N+N*K; 
  block.OutputPort(02).Dimensions = block.DialogPrm(1).Data*(3+K)+1;
  %   3)  mDot, p, T, Xc
  block.OutputPort(03).Dimensions = 3+K;

  %% Register methods
  block.RegBlockMethod('InitializeConditions',    @InitConditions);  
  block.RegBlockMethod('Outputs',                 @Output);  
  block.RegBlockMethod('Derivatives',             @Derivative);  
  %block.RegBlockMethod('SetInputPortSamplingMode',@SetInpPortFrameData);
 
%endfunction

function InitConditions(block)
% Check prequisities
    global psSimStruct
    K = psSimStruct.psThermProp.K;
    connectorUpIsRestriction = block.DialogPrm(9).Data==1;
    connectorDownIsRestriction = block.DialogPrm(10).Data==1;
    if (block.DialogPrm(2).Data(1) ~= 1 && ~connectorUpIsRestriction) || ...
       (block.DialogPrm(2).Data(2) ~= 1 && ~connectorDownIsRestriction) 
       error('psMultiGasPipe:InitConditions',['You cannot have more than ' ...
                           'one input if the corresponding connector is ' ...
                           'not a valve.'])
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Some constant properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    N = block.DialogPrm(1).Data;
    %    x = block.DialogPrm(3).Data;
    %    A = block.DialogPrm(4).Data;
    pInit = block.DialogPrm(5).Data;
    TInit = block.DialogPrm(6).Data;
    XcInit = block.DialogPrm(7).Data;

    % N parts gives N pressures, N temperatures, N*K Xc and
    % N+1 possible speeds (depending on upstreams connectors).
    pIndxs = 1:N;
    TIndxs = (1:N)+N;
    XcIndxs = reshape((1:(N*K))+2*N,K,[]);
    UIndxs = (1:(N+1))+2*N+N*K; % We may possibly have speeds for all
                                % boundaries, including upstreams and
                                % downstream. 

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Initial condiditons
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    block.ContStates.Data(pIndxs) = pInit;
    %disp('Setting pInit = 10 times normal pInit for zone 1')
    %block.ContStates.Data(pIndxs(1)) = pInit*10;
    block.ContStates.Data(TIndxs) = TInit;
    block.ContStates.Data(XcIndxs) = repmat(XcInit,1,N);
    %disp('Setting xResInit = 0.9 for zone 1')
    %block.ContStates.Data(xResIndxs(1)) = 0.9;
    block.ContStates.Data(UIndxs) = 0;
%endfunction

function Output(block)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Some constant properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    global psSimStruct
    K = psSimStruct.psThermProp.K;
    connectorUpIsRestriction = block.DialogPrm(9).Data==1;
    connectorDownIsRestriction = block.DialogPrm(10).Data==1;
    N = block.DialogPrm(1).Data;
    A = block.DialogPrm(4).Data;
    XcInit = block.DialogPrm(7).Data;
    % N parts gives N pressures, N temperatures, N*K Xc and
    % N+1 possible speeds (depending on upstreams connectors).
    pIndxs = 1:N;
    TIndxs = (1:N)+N;
    XcIndxs = reshape((1:(N*K))+2*N,K,[]);
    UIndxs = (1:(N+1))+2*N+N*K; % We may possibly have speeds for all
                                % boundaries, including upstreams and
                                % downstream. 
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Outputs that are always present
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %   Output dimensions: 
    %     1) mDot, p, T, Xc
    %     2)  p=(1:N), T=(1:N)+N, Xc=(1:(N*K))+2*N, U =(1:(N+1))+2*N+N*K; 
    %     3)  mDot, p, T, Xc
    block.OutputPort(01).Data(2:3) = [block.ContStates.Data(pIndxs(1)) ...
                        block.ContStates.Data(TIndxs(1))]; 
    block.OutputPort(01).Data(3+(1:K)) = block.ContStates.Data(XcIndxs(:,1))'; 
    block.OutputPort(03).Data(2:3) = [block.ContStates.Data(pIndxs(end)) ...
                        block.ContStates.Data(TIndxs(end))]; 
    block.OutputPort(03).Data(3+(1:K)) = block.ContStates.Data(XcIndxs(:,end))'; 

    %% Flows up and down if the connectors are not restrictions and we
    %% have to be the restriction ourselfes
    if ~connectorUpIsRestriction
        %Inputs are: mDot,p,T,Xc
        % We want properties on both side of the upper restriction
        pZones    = [block.InputPort(1).data(2);  ...
                     block.ContStates.Data(pIndxs(1))];
        TZones    = [block.InputPort(1).data(3) ; ...
                     block.ContStates.Data(TIndxs(1))];
        XcZones   = [block.InputPort(1).data(3+(1:K))'; ...
                     block.ContStates.Data(XcIndxs(:,1))'];
        RZones  = zeros(2,1);
        cpZones = zeros(2,1);
        cvZones = zeros(2,1);
        hZones  = zeros(2,1);
        for ii = 1:2
            [hZones(ii,1) cpZones(ii,1) cvZones(ii,1) RZones(ii,1)] = ...
           psThermProp(pZones(ii,1),TZones(ii,1),XcZones(ii,:),[1 1 1 1 0 0 0 0 0]);
        end
        if max(cvZones == 0) == 1
            cvZones(find(cvZones==0)) = eps;
        end
        gammaZones = cpZones./cvZones;

        %rho=p/(R*T); 
        if max(RZones.*TZones == 0) ~= 1
            rhoZones=pZones./(RZones.*TZones); 
        else
            rhoZones=1e20.*[1;1];
        end

        % dm = A*u*rho where rho is density of the gas that is passing the
        % boundary (rho = p/RT) u>0 => rho = rho_up etc. rho_up is the N-1
        % first rho and rho_down is the N-1 last of the pipe rhos's
        UPlug = block.ContStates.Data(UIndxs(1));
        rhoPlugFlow =rhoZones(1).*(UPlug>0)+...
            rhoZones(2).*(UPlug<=0);
        dmPlug = A*UPlug.*rhoPlugFlow;
        block.OutputPort(01).Data(1) = dmPlug;
    else
        block.OutputPort(01).Data(1) = 0;
    end
    if ~connectorDownIsRestriction
        %Inputs are: mDot,p,T,phi,phiRes,xRes
        % We want properties on both side of the upper restriction
        pZones    = [block.ContStates.Data(pIndxs(end)); ...
                     block.InputPort(2).data(2)];
        TZones    = [block.ContStates.Data(TIndxs(end)); ...
                     block.InputPort(2).data(3)];
        XcZones   = [block.ContStates.Data(XcIndxs(:,end))';...
                     block.InputPort(2).data(3+(1:K))'];
        RZones  = zeros(2,1);
        cpZones = zeros(2,1);
        cvZones = zeros(2,1);
        hZones  = zeros(2,1);
        for ii = 1:2
            [hZones(ii,1) cpZones(ii,1) cvZones(ii,1) RZones(ii,1)] = ...
           psThermProp(pZones(ii,1),TZones(ii,1),XcZones(ii,:),[1 1 1 1 0 0 0 0 0]);
        end
        if max(cvZones == 0) == 1
            cvZones(find(cvZones==0)) = eps;
        end
        gammaZones = cpZones./cvZones;

        %rho=p/(R*T); 
        if max(RZones.*TZones == 0) ~= 1
            rhoZones=pZones./(RZones.*TZones); 
        else
            rhoZones=1e20.*[1;1];
        end
       
        % dm = A*u*rho where rho is density of the gas that is passing the
        % boundary (rho = p/RT) u>0 => rho = rho_up etc. rho_up is the N-1
        % first rho and rho_down is the N-1 last of the pipe rhos's
        UPlug = block.ContStates.Data(UIndxs(end));
        rhoPlugFlow =rhoZones(1).*(UPlug>0)+...
            rhoZones(2).*(UPlug<=0);
        dmPlug = A*UPlug.*rhoPlugFlow;
        block.OutputPort(03).Data(1) = dmPlug;
    else 
        block.OutputPort(03).Data(1) = 0;
    end
    % Diag port
    block.OutputPort(02).Data = block.ContStates.Data;
%endfunction

function Derivative(block)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Some constant properties
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % External stuff
    global psSimStruct
    K = psSimStruct.psThermProp.K;
    
    % N parts gives N pressures, N temperatures, N*K Xc and
    % N+1 possible speeds (depending on upstreams connectors).
    N = block.DialogPrm(1).Data;
    pIndxs = 1:N;
    TIndxs = (1:N)+N;
    XcIndxs = reshape((1:(N*K))+2*N,K,[])';
    UIndxs = (1:(N+1))+2*N+N*K; % We may possibly have speeds for all
                                % boundaries, including upstreams and
                                % downstream. 
    A = block.DialogPrm(4).Data;
    x = block.DialogPrm(3).Data;
    V = A*x/N;
    kFric = block.DialogPrm(8).Data;
    connectorUpIsRestriction = block.DialogPrm(9).Data==1;
    connectorDownIsRestriction = block.DialogPrm(10).Data==1;
    NbrInSigsUp   = block.DialogPrm(2).Data(1);
    NbrInSigsDown = block.DialogPrm(2).Data(2);    
    XcInIndxsUp     = repmat((0:(NbrInSigsUp-1))*(3+K)+3,K,1)...
        +repmat((1:K)',1,NbrInSigsUp);
    XcInIndxsDown     = repmat((0:(NbrInSigsDown-1))*(3+K)+3,K,1)...
        +repmat((1:K)',1,NbrInSigsDown);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % In signals
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    dmUp          = block.InputPort(1).data((0:(NbrInSigsUp-1))*(3+K)+1);
    pUp           = block.InputPort(1).data((0:(NbrInSigsUp-1))*(3+K)+2);
    TUp           = block.InputPort(1).data((0:(NbrInSigsUp-1))*(3+K)+3);
    XcUp          = block.InputPort(1).data(XcInIndxsUp)';

    dmDown        = -block.InputPort(2).data((0:(NbrInSigsDown-1))*(3+K)+1);
    pDown         =  block.InputPort(2).data((0:(NbrInSigsDown-1))*(3+K)+2);
    TDown         =  block.InputPort(2).data((0:(NbrInSigsDown-1))*(3+K)+3);
    XcDown        =  block.InputPort(2).data(XcInIndxsDown)';

    QDotInTot    = 0; %block.InputPort(xx).data;
    QDotIn = ones(N,1)*QDotInTot/N;

    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Calculation of flows and other variables needed for the derivatives
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Upstreams props
    RUp = zeros(size(pUp));
    for ii = 1:length(RUp)
        [dmy1 dmy2 dmy3 RUp(ii)] = psThermProp(pUp(ii),TUp(ii),XcUp(ii,:),[0 0 0 1 0 0 0 0 0]);
    end
    
    mUp = pUp*V./(RUp.*TUp);
    % Note: This is not the real mass upstreams, the mass is
    % just twice the mass that belongs to the plug part of the outside
    % volumes 
    rhoUp = pUp./(RUp.*TUp); 
    
    % Downstreams props
    RDown = zeros(size(pDown));
    for ii = 1:length(RDown)
        [dmy1 dmy2 dmy3 RDown(ii)] = psThermProp(pDown(ii),TDown(ii),XcDown(ii,:),[0 0 0 1 0 0 0 0 0]);
    end
    mDown = pDown*V./(RDown.*TDown);
    % Note: This is not the real mass upstreams, the mass is
    % just twice the mass that belongs to the plug part of the outside
    % volumes 
    rhoDown = pDown./(RDown.*TDown); 
    % Props for the N zones 
    pZones  = block.ContStates.Data(pIndxs);
    TZones  = block.ContStates.Data(TIndxs);
    XcZones = block.ContStates.Data(XcIndxs')'; % The double ' is because
                                                % of a Matlab convention
                                                % which screws up the 1
                                                % zone case otherwise

    RZones = zeros(N,1);
    for ii = 1:N
        [dmy1 dmy2 dmy3 RZones(ii)] = psThermProp(pZones(ii),TZones(ii), ...
                                                  XcZones(ii,:),[0 0 0 1 0 0 0 0 0 ]);
    end
    mZones = pZones*V./(RZones.*TZones);
    rhoZones=pZones./(RZones.*TZones); 
    
    %    disp(sprintf('In: psMultigasPipe m_Tot = %f',mZones))
    
    % Internal plugs only
    % dm = A*u*rho where rho is density of the gas that is passing the
    % boundary (rho = p/RT) u>0 => rho = rho_up etc. rho_up is the N-1
    % first rho and rho_down is the N-1 last of the pipe rhos's
    if N >= 2
        % Pressure difference over the N-1 internal zone boundaries
        DpIntZones= pZones(1:(end-1))-pZones(2:end); % N-1 first - N-1 last
        UIntPlugs = block.ContStates.Data(UIndxs(2:(end-1)));
        rhoIntPlugFlow =rhoZones(1:(end-1)).*(UIntPlugs>0)+...
            rhoZones(2:end).*(UIntPlugs<=0);
        dmIntPlugs = A*UIntPlugs.*rhoIntPlugFlow; % (Size N-1)
    end
    
    % External plugs
    if ~connectorUpIsRestriction 
        DpUp          = pUp-pZones(1);
        UPlugUp       = block.ContStates.Data(UIndxs(1));
        rhoPlugFlowUp = rhoUp*(UPlugUp>0)+rhoZones(1)*(UPlugUp<=0);
        dmUp      = A*UPlugUp*rhoPlugFlowUp;
    end
    
    if ~connectorDownIsRestriction 
        DpDown          = pZones(end)-pDown;
        UPlugDown       = block.ContStates.Data(UIndxs(end));
        rhoPlugFlowDown = rhoZones(end)*(UPlugDown>0)+rhoDown*(UPlugDown<=0);
        dmDown          = -A*UPlugDown*rhoPlugFlowDown;
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Calculate the derivatives
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Ax = B with constant V (dV=0) eliminates parts of the A-matrix 
    % Use therefore psdX_SigneZone_constV

    XpT  = [pZones; TZones];

    dm = zeros(N,N);
    % Flöde från 2 till 1
    for ii = 1:(N-1)
        dm(ii,ii+1) = -dmIntPlugs(ii);
        dm(ii+1,ii) = -dm(ii,ii+1);
    end
    dQ = QDotIn;

    %    if N == 3 && any(TZones < 298)
    %     disp('Break here')
    % end
    
    
    [dXpT dXgc] = psDX_SingleZone_ConstV(XpT,repmat(V,N,1),XcZones,dm,dQ,TUp,XcUp,dmUp,TDown,XcDown,dmDown);
    
    %  - dUdt calculation (for internal plugs only)    
    if N >= 2
        mPlugsUp   =  mZones(1:(end-1));
        mPlugsDown =  mZones(2:end);
        dudt  =[0 ;2*DpIntZones*A./(mPlugsUp+mPlugsDown)-...
                UIntPlugs*kFric;  0]; % N+1 Vector

    else
        dudt = [0 0];
    end
    
    if ~connectorUpIsRestriction 
        dudt(1) = 2*DpUp*A/(mZones(1)+mUp)-...
                  block.ContStates.Data(UIndxs(1))*kFric;
    end
    if ~connectorDownIsRestriction 
        dudt(end) = 2*DpDown*A/(mZones(end)+mDown)-...
            block.ContStates.Data(UIndxs(end))*kFric;
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Output the derivatives 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % - dpdt output
    if min(pZones) >=0
    %  dX = [dp1 dT1 ... dpN dTN ]
        block.Derivatives.Data(pIndxs)   = dXpT(pIndxs);
    else
        block.Derivatives.Data(pIndxs)   = NaN;
    end
    % - dTdt output
    %  dX = [dp1 dT1 ... dpN dTN ]
    block.Derivatives.Data(TIndxs)   = dXpT(TIndxs);
    
    % - dxResdt output    
    block.Derivatives.Data(XcIndxs)= dXgc;

    % - dUdt output    
    block.Derivatives.Data(UIndxs)   = dudt;

    
    %block.Derivatives(2).Data = 2;
%endfunction

%function SetInpPortFrameData(block, idx, fd)
%  
%  block.InputPort(idx).SamplingMode = fd;
%  block.OutputPort(01).SamplingMode  = fd;
%  block.OutputPort(02).SamplingMode  = fd;
%  block.OutputPort(03).SamplingMode  = fd;
%endfunction

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% DEBUG stuff
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%...
%...
function doNothing()

figure;
subplot(4,1,1)
hold on
plot(psSimOut.time,psSimOut.pge,'g',psSimOut.time,psSimOut.pcb,'r')
subplot(4,1,2)
plot(psSimOut.time,psSimOut.uIV,'g',psSimOut.time,psSimOut.uEV,'r')

subplot(4,1,3)
plot(psSimOut.time,psSimOut.mDotEV,'r',psSimOut.time,psSimOut.mDotIV,'g')
subplot(4,1,4)
plot(psSimOut.time,psSimOut.Tge,psSimOut.time,psSimOut.Tb,psSimOut.time,psSimOut.Tu)